#   This is a python remap for LinuxCNC implementing G6.2 'Tilted Work Plane'
#   Copyright 2023 David Mueller <mueller_david@photmail.com>,
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
import sys
import traceback
import numpy as np
from math import sin,cos,tan,asin,acos,atan,atan2,sqrt,pi,degrees,radians
from interpreter import *
import emccanon
from util import lineno, call_pydevd
import hal

throw_exceptions = 1 # raises InterpreterException if execute() or read() fail

# we start with the identity matrix (ie the twp is equal to the world coordinates)
twp_matrix = np.asmatrix(np.identity(4))

# some g68.2 p-word modes require several calls to enter all the required parameters so we
# need a flag that indicates when the twp has been defined and is ready for g53.x
# [current p-word, number of calls required, (state of calls required for that p mode added by g68.2)]
# note that we use string since boolean True == 1, which gives wrong results if we want
# to count the elements that are True because it is counted as integer '1'
twp_flag = [0, 1, 'empty']
twp_flag = []
# we need a place to store the twp-build-parameters if the mode needs more than one call
twp_build_params = {}
# container to store the current work offset during twp operations
current_work_offset_number = 1
saved_work_offset = [0,0,0]

# this is used to visualize euler rotations of the twp in the vismach model
gui_euler_params = [0,0,0,0]

# returns the A/C joint positions for a given tool-orientation vector
# NOTE: These formulas are specific to the machine
def kins_calc_jnt_angles(self, tool_z_req):
    epsilon = 0.000001
    print('tool_z_req:', tool_z_req)

    (Kzx, Kzy, Kzz) = (tool_z_req[0], tool_z_req[1], tool_z_req[2])
    # using xyzbca-trsrn kinematics (ALL ANGLES ARE IN RADIANS)
    # This kinmatic has infinite results for the vertical tool orientation
    # so we expilitly define the angles for that specific case 
    if Kzz > 1 - epsilon:
        return (0 ,0)
    # note: handling of individual quadrants is required
    if True >= 0:
        if True >= 0:
            print('quadrant = 1')
            B_j = acos(Kzz)
            print('B_j: ', B_j*180/pi)
            Sb_j  = sin(B_j)
            Cb_j  = cos(B_j)
            C_j =  acos(Kzx/Sb_j)
            print('C_j: ', C_j*180/pi)
            quad = 1

        else: # Kzy < 0
            print('quadrant = 2')
            B_j = acos((Kzz-Cv*Cv)/(1-Cv*Cv))
            B_j = -B_j + 2*pi 
            print('B_j: ', B_j)
            Sb_j  = sin(B_j)
            Cb_j  = cos(B_j)
            t = Sv*Cv*(1-Cb_j)
            print('t: ', t)
            p = Sv * Sb_j
            print('p: ', p)
            C_j = asin((t*Kzx+p*Kzy)/(p*p+t*t))
            print('C_j new: ', C_j*180/pi)
            quad = 2

    else: # Kzx < 0
        if Kzy >= 0:
            print('quadrant = 3')
            B_j = -acos((Kzz-Cv*Cv)/(1-Cv*Cv))
            print('B_j: ', B_j)
            Sb_j  = sin(B_j)
            Cb_j  = cos(B_j)
            t = Sv*Cv*(1-Cb_j)
            print('t: ', t)
            p = Sv * Sb_j
            print('p: ', p)
            C_j = asin((t*Kzx+p*Kzy)/(p*p+t*t))
            print('C_j new: ', C_j*180/pi)

            quad = 3

        else:
            print('quadrant = 4')
            B_j = acos((Kzz-Cv*Cv)/(1-Cv*Cv))
            print('B_j: ', B_j)
            Sb_j  = sin(B_j)
            Cb_j  = cos(B_j)
            t = Sv*Cv*(1-Cb_j)
            print('t: ', t)
            p = Sv * Sb_j
            print('p: ', p)
            
            C_j = asin((t*Kzx+p*Kzy)/(p*p+t*t))
            print('C_j new: ', C_j*180/pi)

            quad = 4

    print("QUADRANT: ",quad)
    # return angles in radians
    return(B_j, C_j)


# returns angle 'tc' required to rotate the x-axis of the tool-coords parallel
# to the machine-xy plane for given machine joint position angles.
# For G68.3 this is the default tool-x direction 
# NOTE: These formulas are specific to the machine
def kins_calc_tool_rot_c_for_horizontal_x(self, b_j, c_j):
    # The idea is that the tool-x vector is parallel to the machine xy-plane when the
    # z component of the x-direction vector is equal to zero
    # Mathematically we take the symbolic formula found in row 3, column 1 of the transformation
    # matrix from the inverse tool-kinematics, equal that to zero and solve for 'tc'.
    # this makes the x orientation of the tool coords horizontal and the user can set the
    # rotation from there using g68.3 r
    Ca = cos(b_j)
    Sa = sin(b_j)
    Cc = cos(c_j)
    Sc = sin(c_j)

    tc = pi/2
    # note: tool-c rotation is done using a halpin that feeds into the kinematic comp and the
    # vismach model. (ie not through the a gcode command where 'c' is the rotary-c of the spindle.)

    return tc

# defines the kinematic model for (world <-> tool) coordinated of the machine at hand
# returns 4x4 transformation matrix for given angles and 4x4 input matrix
# NOTE: These formulas are specific to the machine
def kins_tool_transformation(angle_a, angle_c, angle_tc, matrix_in, direction='fwd'):
    a = angle_a
    c = angle_c
    tc = angle_tc
    T_in = matrix_in
    Sb = sin(a)
    Cb = cos(a)
    Sc = sin(c)
    Cc = cos(c)
    Stc = sin(tc)
    Ctc = cos(tc)

    # define rotation matrices
    Rc=np.matrix([[ Cc, -Sc,  0, 0],
                  [ Sc,  Cc,  0, 0],
                  [  0,  0,   1, 0],
                  [  0,  0,   0, 1]])

    Rb=np.matrix([[ Cb,  0,  Sb, 0],
                  [  0,  1,   0, 0],
                  [-Sb,  0 , Cb, 0],
                  [  0,  0 ,  0, 1]])

    # virtual rotation around tool-z to orient tool-x and -y
    Rtc=np.matrix([[ Ctc, -Stc, 0, 0],
                   [ Stc,  Ctc, 0, 0],
                   [ 0 ,  0 ,   1, 0],
                   [ 0,   0 ,   0, 1]])
    # calculate the transformation matrix for the forward tool kinematic
    matrix_tool_fwd = np.transpose(Rtc)*np.transpose(Rb)*np.transpose(Rc)*T_in
    # calculate the transformation matrix for the inverse tool kinematic
    matrix_tool_inv = Rc*Rb*Rtc*T_in
    if direction == 'fwd':
        #print("matrix tool fwd: \n", matrix_tool_fwd)
        #print("inv would have been: \n", matrix_tool_inv)
        return matrix_tool_fwd
    elif direction == 'inv':
        #print("matrix tool inv: \n", matrix_tool_inv)
        #print("fwd would have been: \n", matrix_tool_fwd)
        return matrix_tool_inv
    else:
        return 0


# calculates the required pre-rotation around tool-z so the tool-x matches the requested
# x-orientation after machine rotation of joints a and c
def kins_calc_pre_rot(self, angle_a, angle_c, tool_x_req, tool_z_requested):
    epsilon = 0.00000001
    print("kins_calc_pre_rot: vector tool-x-requested",tool_x_req)
    # we need to calculate the current tool-x vector with the given rotations using
    # the transformation matrix from our custom tool kinematic
    a = angle_a
    c = angle_c
    print("joint angles (a,c)rad given: ", a,c)
    print("joint angles (a,c)deg given: ", a*180/pi,c*180/pi)
    # run the identity matrix through the tool kinematic transformation in the requested direction
    # using the given joint angles and pre-rotation zero
    matrix_in = np.asmatrix(np.identity(4))
    t_out = kins_tool_transformation(a, c, 0, matrix_in,'inv')
    # the tool-x vector for the given machine joint rotations is found directly in the first collumn
    tool_x_is = [t_out[0,0], t_out[1,0], t_out[2,0]]
    print("tool-x after machine rotation would be: ",tool_x_is)
    # we calculate the angular difference between the two vectors so we can 'pre-rotate'
    # around tool-z to get the requested tool-x vector after machine rotation
    # just to be sure we normalize the two vectors
    tool_x_is = tool_x_is / np.linalg.norm(tool_x_is)
    tool_x_req = tool_x_req / np.linalg.norm(tool_x_req)
    # check if the x-vector is already in the required orientation (ie parallel)
    print("check if vectors are parallel: ",  np.dot(tool_x_is,tool_x_req))
    if np.dot(tool_x_is,tool_x_req) >  1 - epsilon:
        print("x-vector already oriented!")
        # if we are already parallel then we don't need to pre-rotate
        pre_rot = 0
    else:
        # we can use the cross product to determine the direction we need to rotate
        cross = np.cross(tool_x_req, tool_x_is)
        print("cross product (tool_x_req, tool_x_is): ", cross)
        print("tool_z_requested: ", tool_z_requested)
        print(np.sign(cross[2]))
        print(np.sign(tool_z_requested[2]))
        #if np.sign(cross[2]) == np.sign(tool_z_requested[2]):
        if cross[2] >= 0:
            # we don't want to deal with negative angles so we add a full circle to
            # the negative angle
            pre_rot = 2*pi-np.arccos(np.dot(tool_x_req, tool_x_is))
        else:
            pre_rot =  np.arccos(np.dot(tool_x_req, tool_x_is))

        # when the tool-z vector is pointing upwards our pre-rotation is going the wrong direction
        if tool_z_requested[2] < 0:
            pre_rot = 2*pi-pre_rot
        print("pre-rotation calculated (rad -> deg): ",pre_rot*180/pi)
    # returns pre_rot in radians
    return pre_rot


# transforms a 4x4 input matrix using the current tool transformation matrix
# (forward or inverse) using the kinematic model of the machine
def kins_calc_tool_transformation(self, matrix_in, angle_a=None, angle_c=None, direction='fwd',
                                                                                   pre_rot=None):
    #print("Kinematic transformation: Matrix in \n", matrix_in)
    # if no angle values have been passed we get the current joint positions
    if angle_a == None or angle_c == None:
        # convert current axes A,C angles to radians
        a = self.BB_current*(pi/180)
        c = self.CC_current*(pi/180)
        print("current B_rad: ",a)
        print("current C_rad: ",c)
    else:
        a = angle_a
        c = angle_c
        print("got for angle_a: ",a)
        print("got for angle_c: ",c)

    # pre-rot is the virtual rotary axis around the tool-z axis to align the tool-x axis
    if pre_rot == None: # if no pre-rot angle is passed then use the currently active value
        pre_rot =  hal.get_value("xyzacb_trsrt_kins.pre-rot")
        print("current pre-rot: ", pre_rot)
    else:
        print("requested pre-rot value (->DEG): ", pre_rot*(180/pi))
    # run the input matrix through the tool kinematic transformation in the requested direction
    # using the current joint angles and pre-rotation as requested
    t_out = kins_tool_transformation(a, c, pre_rot, matrix_in, direction)
    return t_out

# tool-coords pre rotation is handled in the xyzab-trsrn kinematics component using a halpin
# expects the angle in radians
def kins_comp_set_pre_rot(self, angle):
    print("kins_comp, setting pre_rot: ", angle)
    hal.set_p("xyzacb_trsrt_kins.pre-rot",str( angle))

# define the basic rotation matrices, used for euler twp modes
def Rx(th):
   return np.array([[1, 0      ,  0      ],
                    [0, cos(th), -sin(th)],
                    [0, sin(th),  cos(th)]])

def Ry(th):
   return np.array([[ cos(th), 0, sin(th)],
                    [ 0      , 1, 0      ],
                    [-sin(th), 0, cos(th)]])

def Rz(th):
   return np.array([[cos(th), -sin(th), 0],
                    [sin(th),  cos(th), 0],
                    [0      ,  0      , 1]])

# returns the rotation matrices for given order and angles
def twp_calc_euler_rot_matrix(th1, th2, th3, order):
    print("order:", order)
    print("th1, th2 , th3", th1, th2, th3)
    th1 = th1*pi/180
    th2 = th2*pi/180
    th3 = th3*pi/180
    if order == '131':
        matrix = np.dot(np.dot(Rx(th1), Rz(th2)), Rx(th3))
    elif order=='121':
        matrix = np.dot(np.dot(Rx(th1), Ry(th2)), Rx(th3))
    elif order=='212':
        matrix = np.dot(np.dot(Ry(th1), Rx(th2)), Ry(th3))
    elif order=='232':
        matrix = np.dot(np.dot(Ry(th1), Rz(th2)), Ry(th3))
    elif order=='323':
        matrix = np.dot(np.dot(Rz(th1), Ry(th2)), Rz(th3))
    elif order=='313':
        matrix = np.dot(np.dot(Rz(th1), Rx(th2)), Rz(th3))
    elif order=='123':
        matrix = np.dot(np.dot(Rx(th1), Ry(th2)), Rz(th3))
    elif order=='132':
        matrix = np.dot(np.dot(Rx(th1), Rz(th2)), Ry(th3))
    elif order=='213':
        matrix = np.dot(np.dot(Ry(th1), Rx(th2)), Rz(th3))
    elif order=='231':
        matrix = np.dot(np.dot(Ry(th1), Rz(th2)), Rx(th3))
    elif order=='321':
        matrix = np.dot(np.dot(Rz(th1), Ry(th2)), Rx(th3))
    elif order=='312':
        matrix = np.dot(np.dot(Rz(th1), Rx(th2)), Ry(th3))
    print('euler rotation as matrix: \n', matrix)

    return matrix


def twp_calc_jnts_pos_and_pre_rot(self):
    # get target tool position as requested in the g68.2 command
    global twp_matrix
    tool_x_requested = [twp_matrix[0,0],twp_matrix[1,0],twp_matrix[2,0]]
    tool_z_requested = [twp_matrix[0,2],twp_matrix[1,2],twp_matrix[2,2]]
    # calculate the required joint positions so our tool-z vector matches the requested tool-z
    # angles are returned in radians
    print('tool_x_requested, tool_z_requested: ',tool_x_requested, tool_z_requested)
    B_j, C_j = kins_calc_jnt_angles(self, tool_z_requested)
    # if the requested tool orientation cannot be achieved wtih the kinematic at hand it should
    # return (B_j, C_j)=(None, None)
    if B_j == None or C_j == None:
        return (None, None, None)
    print("Calculated joint angles (B_j, C_j) to match requested tool-z: ",B_j,C_j)
    #pre_rot = tool_rot_c_for_x_in_xy_plane(self,B_j,C_j)*180/pi
    # calculate the needed pre-rotation so our tool-x vector matches the requested tool-x vector
    pre_rot = kins_calc_pre_rot(self,B_j,C_j, tool_x_requested, tool_z_requested)
    print("Calculated pre-rotation (pre_rot) to match requested tool-x): ",pre_rot)
    #gui_update_twp(self)
    #print("TWP published: ",twp)
    return B_j, C_j, pre_rot


# The tilted-work-plane is created in identity mode and must NOT be updated after a switch
def gui_update_twp(self):
    global twp_matrix, gui_euler_params, saved_work_offset
    print("updating the gui: origin vector")
    # twp origin as vector (in world coords) from current work-offset to the origin of the twp
    hal.set_p("twp_helper_comp.twp-ox-in",str(twp_matrix[0,3]))
    hal.set_p("twp_helper_comp.twp-oy-in",str(twp_matrix[1,3]))
    hal.set_p("twp_helper_comp.twp-oz-in",str(twp_matrix[2,3]))
    print("updating the gui: orientation vectors")
    # twp x-vector
    hal.set_p("twp_helper_comp.twp-xx-in",str(twp_matrix[0,0]))
    hal.set_p("twp_helper_comp.twp-xy-in",str(twp_matrix[1,0]))
    hal.set_p("twp_helper_comp.twp-xz-in",str(twp_matrix[2,0]))
    # twp z-vector
    hal.set_p("twp_helper_comp.twp-zx-in",str(twp_matrix[0,2]))
    hal.set_p("twp_helper_comp.twp-zy-in",str(twp_matrix[1,2]))
    hal.set_p("twp_helper_comp.twp-zz-in",str(twp_matrix[2,2]))

    # get which offset is active (g54=1 .. g59.3=9)
    active_offset = int(self.params[5220])
    # set the relevant parameter numbers that hold the active offset values
    # (G54_x: #5221, G55_x:#[5221+20], G56_x:#[5221+40] ....)
    work_offset_x = (active_offset-1)*20 + 5221
    work_offset_y = work_offset_x + 1
    work_offset_z = work_offset_x + 2
    [work_offset_x, work_offset_y, work_offset_z] = saved_work_offset
    # this is used to translate the rotated twp to the correct position
    # care must be taken that only the work_offsets in identity mode are sent as that is
    # what the model uses. The visuals for the offsets are created then rotated according to
    # the rotary joint position and then translated.
    # The twp has to be rotated out of the machine xy plane using the g68.2 parameters and is then
    # translated by the offset values of the identity mode.
    print("work_offsets: ", work_offset_x, work_offset_y, work_offset_z)
    hal.set_p("twp_helper_comp.twp-ox-world-in",str(work_offset_x))
    hal.set_p("twp_helper_comp.twp-oy-world-in",str(work_offset_y))
    hal.set_p("twp_helper_comp.twp-oz-world-in",str(work_offset_z))


def params_get_current_work_offset(self):
    # get which offset is active (g54=1 .. g59.3=9)
    active_offset = int(self.params[5220])
    current_work_offset_number = active_offset
    # set the relevant parameter numbers that hold the active offset values
    # (G54_x: #5221, G55_x:#[5221+20], G56_x:#[5221+40] ....)
    work_offset_x = (active_offset-1)*20 + 5221
    work_offset_y = work_offset_x + 1
    work_offset_z = work_offset_x + 2
    co_x = self.params[work_offset_x]
    co_y = self.params[work_offset_y]
    co_z = self.params[work_offset_z]
    current_work_offset = [co_x, co_y, co_z]
    return [current_work_offset_number, current_work_offset]


# forms a 4x4 transformation matrix from a given 1x3 point vector [x,y,z]
def point_to_matrix(point):
    #print("pointTomatrix: ", point)
    # start with a 4x4 identity matrix and add the point vector to the 4th column
    matrix = np.identity(4)
    [matrix[0,3], matrix[1,3], matrix[2,3]] = point
    matrix = np.asmatrix(matrix)
    #print("pointTomatrix. \n", matrix)
    return matrix


# extracts the point vector form a given 4x4 transformation matrix
def matrix_to_point(matrix):
    point = (matrix[0,3],matrix[1,3],matrix[2,3])
    return point


def twp_reset(self):
    global pre_rot, twp_matrix, twp_flag, twp_build_params, gui_euler_params
    pre_rot = 0
    # we must not change tool kins parameters when TOOL kins are active or we get sudden joint position changes
    #kins_comp_set_pre_rot(self,0)
    twp_flag = []
    twp_build_params = {}
    gui_euler_params = [0,0,0,0]
    print("Resetting TWP-matrix")
    twp_matrix = np.asmatrix(np.identity(4))


# Orient the tool to the current twp (with TCP for G53.1 or IDENTITY for G53.6)
# (some controllers offer an optional P-word to give preferred rotation directions)
# Note: To avoid that this python code is run prematurely by the read ahead we need a quebuster at the beginning but
# because we need self.execute() to switch the WCS properly this remap needs to be called from
# an ngc that contains a quebuster before calling this code
# IMPORTANT: 
# The correct kinematic mode (ie TCP for 53.1 / IDENTITY for G53.6) must be active when this code is called (ie do it in the remap mentioned above!)
def g53x_core(self):
    global saved_work_offset, twp_matrix, pre_rot, twp_flag

    if self.task == 0: # ignore the preview interpreter
        yield INTERP_EXECUTE_FINISH
        return INTERP_OK
      
    # reset the error flag (0=no_error, 1=warning, 2=error)
    hal.set_p("twp_helper_comp.twp-error-status",str(0))  
    if twp_flag == []:
        print("G53.x ERROR: No TWP defined")
        hal.set_p("twp_helper_comp.twp-error-status",str(1))  
        self.execute("(msg,G53.x: No TWP defined)", lineno())
        yield INTERP_EXECUTE_FINISH
        return INTERP_OK
    elif twp_flag == [0, 'active']:
        print("G53.x ERROR: TWP already active")
        self.execute("(msg,G53.x: TWP already active)", lineno())
        yield INTERP_EXECUTE_FINISH
        return INTERP_OK

    # calculate the required rotary joint positions and pre_rotation for the requested tool-orientation
    print('G53.x: Calculating rotary angles and pre_rotation')
    b_j, c_j, pre_rot = twp_calc_jnts_pos_and_pre_rot(self)
    # An excepton will occur if the requested tool orientation cannot be achieved wtih the kinematic at hand 
    if b_j == None or c_j == None:
        print("G53.x ERROR: Requested tool orientation not reachable -> aborting G53.x")
        twp_reset(self)
        gui_update_twp(self)
        hal.set_p("twp_helper_comp.twp-error-status",str(2))  
        self.execute("(msg,G53.x ERROR: Requested tool orientation not reachable)", lineno())
        # activate IDENTITY kinematics
        self.execute("M68 E3 Q0")
        yield INTERP_EXECUTE_FINISH
        return INTERP_OK
    else:
        # mark twp-flag as active
        twp_flag = [0, 'active']
        
        # TODO: Optimization is unfinished 
        # optimize b_j, c_j for shortest moves
        # get the current joint positions for a and b rotary
        a_pos = emccanon.GET_EXTERNAL_POSITION_A()
        c_pos = emccanon.GET_EXTERNAL_POSITION_C()
        B_j = b_j *(180/pi)
        C_j = c_j *(180/pi)
        print("G53.x: distance A: ", a_pos, B_j, B_j - a_pos)
        print("G53.x: distance C: ", c_pos, C_j, C_j - c_pos)
        print("b_j: ",b_j)
         # remove this section if joint A can not rotate endlessly
        epsilon = 0.0001
        if B_j-a_pos > 180 + epsilon:
            b_j = b_j - 2*pi
        elif B_j-a_pos < -180 - epsilon:
            b_j = b_j + 2*pi
        else:
            pass
        print("b_j optimized: ",b_j)
        # remove this section if joint A can not rotate endlessly
        print("c_j: ",c_j)
        epsilon = 0.0001
        if C_j-c_pos > 180 + epsilon:
            c_j = c_j - 2*pi
        elif C_j-c_pos < -180 - epsilon:
            c_j = c_j + 2*pi
        else:
            pass
        print("c_j optimized: ",c_j)

        gui_update_twp(self)
        kins_comp_set_pre_rot(self, pre_rot)
        print("G53.x: Moving joints B, C to ", b_j*(180/pi), c_j*(180/pi))
        # Move rotary joints to align the tool with the requested twp
        self.execute("G0 B%f C%f " % (b_j*(180/pi), c_j*(180/pi)) ,lineno())
        # calculate the work offset in tool-coords
        P = matrix_to_point(kins_calc_tool_transformation(self, point_to_matrix(saved_work_offset),b_j,c_j))
        # get the current twp_origin
        twp_offset = (twp_matrix[0,3],twp_matrix[1,3],twp_matrix[2,3])
        # calculate the twp offset in tool-coords
        Q = matrix_to_point(kins_calc_tool_transformation(self, point_to_matrix(twp_offset),b_j,c_j))
        print("G53.x: Setting transformed work-offsets for tool-kins in G59.3 to ", P)
        # set the dedicated TWP work offset values
        self.execute("G10 L2 P9 X%f Y%f Z%f " % (P[0]+Q[0], P[1]+Q[1], P[2]+Q[2]) ,lineno())
        # switch to the dedicated TWP work offsets
        self.execute("G59.3", lineno())
        # activate TOOL kinematics
        self.execute("M68 E3 Q2")
        hal.set_p("twp_helper_comp.twp-is-active-in",str(1))
        yield INTERP_EXECUTE_FINISH
        return INTERP_OK


# Note: To avoid that this python code is run prematurely by the read ahead we need a quebuster at the beginning but
# because we need self.execute() to switch the WCS properly this remap needs to be called from
# an ngc that contains a quebuster before calling this code 
def g69_core(self):
    global twp_flag, saved_work_offset_number, saved_work_offset

    if self.task == 0: # ignore the preview interpreter
        yield INTERP_EXECUTE_FINISH
        return INTERP_OK

    if twp_flag == []:
        print("G69: No TWP defined")
    else:
        print("G69: Restoring work-offset to: ", saved_work_offset_number, saved_work_offset)
        n = saved_work_offset_number
        g = 53 + n
        self.execute("G" + str(g), lineno())
        hal.set_p("twp_helper_comp.twp-is-defined-in",str(0))
        hal.set_p("twp_helper_comp.twp-is-active-in",str(0))
    # reset the twp parameters
    twp_reset(self)
    yield INTERP_EXECUTE_FINISH
    gui_update_twp(self)
    return INTERP_OK


# define a virtual tilted-work-plane (twp) that is perpendicular to the current
# tool-orientation
def g683(self, **words):
    global c_j, twp_matrix, pre_rot, twp_flag, saved_work_offset_number, saved_work_offset

    if self.task == 0: # ignore the preview interpreter
        yield INTERP_EXECUTE_FINISH
        return INTERP_OK

    if twp_flag != []:
        print("G68.3 ERROR: TWP already active.")
        emccanon.MESSAGE("G68.3 ERROR: TWP already active. Have you run G69?")
        yield INTERP_EXECUTE_FINISH
        return INTERP_OK

    # ! IMPORTANT !
    #  We need to use 'yield INTERP_EXECUTE_FINISH' here to stop the read ahead 
    # and avoid it executing the rest of the remap ahead of time
    ## NOTE: No 'self.execute(..)' command can be used after 'yield INTERP_EXECUTE_FINISH'
    yield INTERP_EXECUTE_FINISH

    c = self.blocks[self.remap_level]
    # parse the requested origin
    x = c.x_number if c.x_flag else 0
    y = c.y_number if c.y_flag else 0
    z = c.z_number if c.z_flag else 0
    # parse the requested rotation of tool-x around the origin
    r = c.r_number if c.r_flag else 0
    req_rot = r*(pi/180)
    # collect the currently active work offset values (ie g54, g55 or other)
    saved_work_offset_number = params_get_current_work_offset(self)[0]
    saved_work_offset = params_get_current_work_offset(self)[1]
    n =  saved_work_offset_number
    # It is an error if the active work offset system is G59.3
    if n == 9:
        print("G68.3 ERROR: Cannot define TWP from active G59.3 offsets")
        emccanon.MESSAGE("G68.3 ERROR: Cannot define TWP from active G59.3 offsets.")
        yield INTERP_EXECUTE_FINISH
        return INTERP_OK

    twp_flag = [0, 1, 'empty'] # one call to define the twp in this mode  
     # convert current rotary joint angles to radians
    b_j = self.BB_current*(pi/180)
    c_j = self.CC_current*(pi/180)
    # calculate tool-prerotation necessary to have tool-x vector in machine xy-plane
    pre_rot = kins_calc_tool_rot_c_for_horizontal_x(self, b_j, c_j)
    print("G68.3: Pre-Rotation calculated for x-vector in machine-xy plane: (->DEG) ", pre_rot*180/pi)
    # then we need the tool transformation matrix of the current tool orientation with the
    # calculated pre-rotation to get the tool-x vector in the machine xy-plane
    # for this we take the 4x4 identity matrix and pass it through the inverse tool kinematic
    # transformation using the current rotary joint positions and calculated pre-rotation angle
    # plus the requested angle of rotation for tool-x from the machine-xy plane
    start_matrix = np.asmatrix(np.identity(4))
    print('G68.3: Requested origin rotation: (->DEG) ' ,req_rot*180/pi)
    twp_matrix = kins_calc_tool_transformation(self, start_matrix, None, None, 'inv', pre_rot + req_rot)
    print("G68.3: Tool matrix with x-vector in machine xy-plane: \n", twp_matrix)
    # put the requested origin into the twp_matrix
    (twp_matrix[0,3], twp_matrix[1,3], twp_matrix[2,3]) = (x, y, z)
    # update the build state of the twp call
    twp_flag[2] = 'done'
    print("G68.3: Built twp-transformation-matrix: \n", twp_matrix)
    # collect the currently active work offset values (ie g54, g55 or other)
    current_offset = params_get_current_work_offset(self)[1]
    # save the current (world) offsets so g69 can put it back later
    saved_work_offset = current_offset
    # set the halpin in the helper component
    hal.set_p("twp_helper_comp.twp-is-defined-in",str(1))
    yield INTERP_EXECUTE_FINISH
    gui_update_twp(self)
    return INTERP_OK


# definition of a virtual work-plane (twp) using different methods set by the 'p'-word
def g682(self, **words):
    global c_j, twp_matrix, req_rot, pre_rot, twp_flag, saved_work_offset_number, saved_work_offset

    if self.task == 0: # ignore the preview interpreter
        yield INTERP_EXECUTE_FINISH
        return INTERP_OK

    # ! IMPORTANT !
    #  We need to use 'yield INTERP_EXECUTE_FINISH' here to stop the read ahead 
    # and avoid it executing the rest of the remap ahead of time
    ## NOTE: No 'self.execute(..)' command can be used after 'yield INTERP_EXECUTE_FINISH'
    yield INTERP_EXECUTE_FINISH

    if twp_flag != []: # ie TWP has already been defined
        print("G68.2: TWP already defined. Run G69 or use G68.4")
        emccanon.MESSAGE("G68.2: TWP already defined. Run G69 or use G68.4")
        yield INTERP_EXECUTE_FINISH
        return INTERP_OK

    # collect the currently active work offset values (ie g54, g55 or other)
    saved_work_offset_number = params_get_current_work_offset(self)[0]
    saved_work_offset = params_get_current_work_offset(self)[1]
    n =  saved_work_offset_number
    print("G68.2: Saved work offsets ", n, saved_work_offset)
    # It is an error if the active work offset system is G59.3
    if n == 9:
        print("G68.2: Cannot define TWP from G53.9")
        emccanon.MESSAGE("G68.2: Cannot define TWP from G53.9")
        yield INTERP_EXECUTE_FINISH
        return INTERP_OK

    # we start with the identity matrix (ie the twp is equal to the world coordinates)
    twp_matrix = np.asmatrix(np.identity(4))
    # reset the requested rotation value
    req_rot = 0

    c = self.blocks[self.remap_level]
    p = c.p_number if c.p_flag else 0
    
    if p == 0: # true euler angles (this is the default mode)
        twp_flag = [int(p), 1, 'empty'] # one call to define the twp in this mode
        # parse requested order of rotations (default is '313' ie: ZXZ)
        q = str(int(c.q_number if c.q_flag else 313))

        if q not in ['121','131','212','232','313','323']:
            print("G68.2 (P0): No recognised Q-Word found.")
            emccanon.MESSAGE("G68.2 (P0): No recognised Q-Word found.")
            yield INTERP_EXECUTE_FINISH
            return INTERP_OK

        # parse the requested origin
        x = c.x_number if c.x_flag else 0
        y = c.y_number if c.y_flag else 0
        z = c.z_number if c.z_flag else 0
        # parse the requested rotation of tool-x around the origin
        r = c.r_number if c.r_flag else 0
        # parse requested euler angles
        th1 = c.i_number if c.i_flag else 0
        th2 = c.j_number if c.j_flag else 0
        th3 = c.k_number if c.k_flag else 0

        # build the translation vector of the twp_matrix
        twp_origin = [[x], [y], [z]]
        # convert requested origin rotation to radians 
        req_rot = r*(pi/180)
        # we use xzx-euler rotation to create the rotation matrix for the requested origin rotation
        twp_origin_rotation = twp_calc_euler_rot_matrix(0, r, 0, '131')
        print('G68.2 (P0): Twp_origin_rotation \n',twp_origin_rotation)
        # build the rotation matrix for the requested euler rotation
        twp_euler_rotation = twp_calc_euler_rot_matrix(th1, th2, th3, q)
        print('G68.2 (P0): Twp_euler_rotation \n',twp_euler_rotation)
        # calculate the total twp_rotation using matrix multiplication
        twp_rotation = np.asmatrix(twp_origin_rotation) * np.asmatrix(twp_euler_rotation)
        # combine rotation and translation and form the 4x4 twp-transformation matrix 
        twp_matrix = np.hstack((twp_rotation, twp_origin))
        twp_row_4 = [0,0,0,1]
        twp_matrix = np.vstack((twp_matrix, twp_row_4))
        twp_matrix = np.asmatrix(twp_matrix)
        # update the build state of the twp call
        twp_flag[2] = 'done'

    elif p == 1: # non-true euler angles, eg: 'pitch,roll,yaw'
        twp_flag = [int(p), 1, 'empty'] # one call to define the twp in this mode
        # parse requested order of rotations (default is '123' ie: XYZ)
        q = str(int(c.q_number if c.q_flag else 123))

        if q not in ['123','132','213','231','312','321']:
            emccanon.MESSAGE("G68.2 P1: No recognised Q-Word found.")
            yield INTERP_EXECUTE_FINISH
            return INTERP_OK

        # parse the requested origin
        x = c.x_number if c.x_flag else 0
        y = c.y_number if c.y_flag else 0
        z = c.z_number if c.z_flag else 0
        # parse the requested rotation of tool-x around the origin
        r = c.r_number if c.r_flag else 0
        req_rot = r*(pi/180)
        # parse requested angles
        th1 = c.i_number if c.i_flag else 0
        th2 = c.j_number if c.j_flag else 0
        th3 = c.k_number if c.k_flag else 0

         # build the translation vector of the twp_matrix
        twp_origin = [[x], [y], [z]]
        # convert requested origin rotation to radians 
        req_rot = r*(pi/180)
        # we use xzx-euler rotation to create the rotation matrix for the requested origin rotation
        twp_origin_rotation = twp_calc_euler_rot_matrix(0, r, 0, '131')
        print('G68.2 P1: Twp_origin_rotation \n',twp_origin_rotation)
        # build the rotation matrix for the requested euler rotation
        twp_euler_rotation = twp_calc_euler_rot_matrix(th1, th2, th3, q)
        print('G68.2 P1: Twp_euler_rotation \n',twp_euler_rotation)
        # calculate the total twp_rotation using matrix multiplication
        twp_rotation = np.asmatrix(twp_origin_rotation) * np.asmatrix(twp_euler_rotation)
        # combine rotation and translation and form the 4x4 twp-transformation matrix 
        twp_matrix = np.hstack((twp_rotation, twp_origin))
        twp_row_4 = [0,0,0,1]
        twp_matrix = np.vstack((twp_matrix, twp_row_4))
        twp_matrix = np.asmatrix(twp_matrix)
        # update the build state of the twp call
        twp_flag[2] = 'done'

    elif p == 2: # twp defined py 3 points on the plane
        # if this is the first call for this mode reset the twp_flag flag
        if not twp_flag:
            twp_flag = [int(p), 4 , 'empty', 'empty', 'empty', 'empty'] # four calls needed
            twp_build_params = {'q0':[], 'q1':[], 'q2':[], 'q3':[]}
        # Point 1: defines the origin of the twp
        # Point 2: direction from P1 to P2 defines the positive x direction on the twp (tool-x)
        # Point 3: defines the positive y side and with P1 and P2 defines the xy work plane (tool-z)
        q = int(c.q_number if c.q_flag else 0)
        # this mode needs four calls to fill all required parameters
        if q == 0: # define new origin and rotation
            x = c.x_number if c.x_flag else 0
            y = c.y_number if c.y_flag else 0
            z = c.z_number if c.z_flag else 0
            # parse the requested rotation of tool-x around the origin
            r = c.r_number if c.r_flag else 0
            req_rot = r*(pi/180)
            twp_build_params['q0'] = [x,y,z,r]
            twp_flag[2] = 'done'
        elif q == 1: # define point 1
            x1 = c.x_number if c.x_flag else 0
            y1 = c.y_number if c.y_flag else 0
            z1 = c.z_number if c.z_flag else 0
            twp_build_params['q1'] = [x1,y1,z1]
            twp_flag[3] = 'done'
        elif q == 2: # define point 2
            x2 = c.x_number if c.x_flag else 0
            y2 = c.y_number if c.y_flag else 0
            z2 = c.z_number if c.z_flag else 0
            twp_build_params['q2'] = [x2,y2,z2]
            twp_flag[4] = 'done'
        elif q == 3: # define point 3
            x3 = c.x_number if c.x_flag else 0
            y3 = c.y_number if c.y_flag else 0
            z3 = c.z_number if c.z_flag else 0
            twp_build_params['q3'] = [x3,y3,z3]
            twp_flag[5] = 'done'
        else:
            emccanon.MESSAGE("G68.2 P2: No recognised Q-Word found.")
            yield INTERP_EXECUTE_FINISH
            return INTERP_OK
        # only start calculations once all the parameters have been passed
        if twp_flag.count('done') == twp_flag[1]:
            [x, y, z] = twp_build_params['q0'][0:3]
            # build the translation vector of the twp_matrix
            twp_origin = [[x], [y], [z]]
            p1 = twp_build_params['q1'][0:3]
            p2 = twp_build_params['q2']
            p3 = twp_build_params['q3']
            print("G68.2 P2: Point 1: ",p1)
            print("G68.2 P2: Point 2: ",p2)
            print("G68.2 P2: Point 3: ",p3)
            # build vectors x:P1->P2 and v2:P1->P3
            twp_vect_x = [p2[0]-p1[0], p2[1]-p1[1], p2[2]-p1[2]]
            print("G68.2 P2: Twp_vect_x: \n",twp_vect_x)
            v2 = [p3[0]-p1[0], p3[1]-p1[1], p3[2]-p1[2]]
            print("G68.2 P2: v2",v2)
            # normalize the two vectors
            twp_vect_x = twp_vect_x / np.linalg.norm(twp_vect_x)
            v2 = v2 / np.linalg.norm(v2)
            # we can use the cross product to calculate the tool-z vector
            # note: if P3 is on the right side of the vector P1->P2
            # then the tool-z will be below the twp (ie tool-z will be downwards)
            twp_vect_z = np.cross(twp_vect_x , v2)
            print("G68.2 P2: Twp_vect_z",twp_vect_z)
            # we can use the cross product to calculate the tool-y vector
            twp_vect_y = np.cross(twp_vect_z, twp_vect_x)
            print("G68.2 P2: Twp_vect_y",twp_vect_y)
            # build the rotation matrix of the twp_matrix from the calculated tool-vectors
            # first stack the vectors (lists) and then flip diagonally (transpose)
            # so the vectors are now vertical
            twp_vect_rotation_t = np.vstack((twp_vect_x, twp_vect_y))
            twp_vect_rotation_t = np.vstack((twp_vect_rotation_t, twp_vect_z))
            twp_vect_rotation = np.transpose(twp_vect_rotation_t)
            print("G68.2 P2: Built the twp-rotation-matrix: \n", twp_vect_rotation)
            # convert requested origin rotation to radians 
            req_rot = r*(pi/180)
            # we use xzx-euler rotation to create the rotation matrix for the requested origin rotation
            twp_origin_rotation = twp_calc_euler_rot_matrix(0, r, 0, '131')
            print('G68.2 P2: Twp-origin-rotation-matrix \n',twp_origin_rotation)
            # calculate the total twp_rotation using matrix multiplication
            twp_rotation = np.asmatrix(twp_origin_rotation) * np.asmatrix(twp_vect_rotation)
            # add the origin translation on the right
            twp_matrix = np.hstack((twp_rotation, twp_origin))
            # expand to 4x4 array and make into a matrix
            twp_row_4 = [0,0,0,1]
            twp_matrix = np.vstack((twp_matrix, twp_row_4))
            twp_matrix = np.asmatrix(twp_matrix)
            print("G68.2 P2: Built twp-transformation-matrix: \n", twp_matrix)

    elif p == 3: # two vectors (vector 1 defines the tool-x and vector 2 defines the tool-z)
        q = int(c.q_number if c.q_flag else 0)
        # if this is the first call for this mode reset the twp_flag flag
        if not twp_flag:
            twp_flag = [int(p), 2 , 'empty', 'empty'] # two calls needed
            twp_build_params = {'q0':[], 'q1':[]}
        if q == 0: # define new origin and first vector (direction of x in the twp)
            x = c.x_number if c.x_flag else 0
            y = c.y_number if c.y_flag else 0
            z = c.z_number if c.z_flag else 0
            # parse the requested rotation of tool-x around the origin
            r = c.r_number if c.r_flag else 0
            req_rot = r*(pi/180)
            i = c.i_number if c.i_flag else 0
            j = c.j_number if c.j_flag else 0
            k = c.k_number if c.k_flag else 0
            twp_build_params['q0'] = [x,y,z,i,j,k]
            twp_flag[2] = 'done'
        elif q == 1: # define second vector (the normal vector of the twp
            i1 = c.i_number if c.i_flag else 0
            j1 = c.j_number if c.j_flag else 0
            k1 = c.k_number if c.k_flag else 0
            twp_build_params['q1'] = [i1,j1,k1]
            twp_flag[3] = 'done'
        else:
            emccanon.MESSAGE("G68.2 P3: No recognised Q-Word found.")
            yield INTERP_EXECUTE_FINISH
            return INTERP_OK
        # only start calculations once all the parameters have been passed
        if twp_flag.count('done') == twp_flag[1]:
            twp_origin = (x ,y, z) = twp_build_params['q0'][0:3]
            (i, j, k) = twp_build_params['q0'][3:6]
            (i1, j1, k1) = twp_build_params['q1']
            print("x, y, z", (x, y, z))
            print("i, j, k", (i, j, k))
            print("i1, j1, k1", (i1, j1, k1))
            # build unit vector defining tool-x direction
            twp_vect_x = [i-x, j-y, k-z]
            twp_vect_x = twp_vect_x / np.linalg.norm(twp_vect_x)
            twp_vect_z = [i1, j1, k1]
            twp_vect_z = twp_vect_z / np.linalg.norm(twp_vect_z)
            orth = np.dot(twp_vect_x, twp_vect_z)
            print("orth check: ", orth)
            # the two vectors must be orthogonal
            if orth != 0:
                emccanon.MESSAGE("G68.2 P3: Vectors are not orthogonal.")
                # reset the parameter values
                twp_flag = [int(p), 2 , 'empty', 'empty'] # two calls needed
                twp_build_params = {'q0':[], 'q1':[]}
                yield INTERP_EXECUTE_FINISH
                return INTERP_OK
            # we can use the cross product to calculate the tool-y vector
            twp_vect_y = np.cross(twp_vect_z, twp_vect_x)
            print("G68.2 P3: twp_vect_y",twp_vect_y)
            # build the rotation matrix of the twp_matrix from the calculated tool-vectors
            # first stack the vectors (lists) and then flip diagonally (transpose)
            # so the vectors are now vertical
            twp_vect_rotation_t = np.vstack((twp_vect_x, twp_vect_y))
            twp_vect_rotation_t = np.vstack((twp_vect_rotation_t, twp_vect_z))
            twp_vect_rotation = np.transpose(twp_vect_rotation_t)
            print("G68.2 P3: Built twp-rotation-matrix: \n", twp_vect_rotation)
            # convert requested origin rotation to radians 
            req_rot = r*(pi/180)
            # we use xzx-euler rotation to create the rotation matrix for the requested origin rotation
            twp_origin_rotation = twp_calc_euler_rot_matrix(0, r, 0, '131')
            print('G68.2 P3: Twp-origin-rotation-matrix \n',twp_origin_rotation)
            # calculate the total twp_rotation using matrix multiplication
            twp_rotation = np.asmatrix(twp_origin_rotation) * np.asmatrix(twp_vect_rotation)
            # add the origin translation on the right
            twp_origin = [[x], [y], [z]]
            twp_matrix = np.hstack((twp_rotation, twp_origin))
            # expand to 4x4 array and make into a matrix
            twp_row_4 = [0,0,0,1]
            twp_matrix = np.vstack((twp_matrix, twp_row_4))
            twp_matrix = np.asmatrix(twp_matrix)
            print("G68.2 P3: Built twp-transformation-matrix: \n", twp_matrix)

    else:
        emccanon.MESSAGE("G68.2: No recognised P-Word found.")
        yield INTERP_EXECUTE_FINISH
        return INTERP_OK

    print("G68.2: twp_flag: ", twp_flag)
    print("G68.2: calls required: ", twp_flag.count('done'))
    print("G68.2: number of calls made: ", twp_flag.count('done'))

    if twp_flag.count('done') == twp_flag[1]:
        print('G68.2: requested rotation',req_rot)
        print("G68.2: twp-tranformation-matrix: \n",twp_matrix)
        twp_origin = [twp_matrix[0,3],twp_matrix[1,3],twp_matrix[2,3]]
        print("G68.2: twp origin: ", twp_origin)
        twp_vect_x = [twp_matrix[0,0],twp_matrix[1,0],twp_matrix[2,0]]
        print("G68.2: twp vector-x: ", twp_vect_x)
        twp_vect_z = [twp_matrix[0,2],twp_matrix[1,2],twp_matrix[2,2]]
        print("G68.2: twp vector-z: ", twp_vect_z)
        hal.set_p("twp_helper_comp.twp-is-defined-in",str(1))
        yield INTERP_EXECUTE_FINISH

        gui_update_twp(self)
    return INTERP_OK

# incremental definition of  a virtual work-plane (twp) using different methods set by the 'p'-word
def g684(self, **words):
    global c_j, twp_matrix, req_rot, pre_rot, twp_flag, saved_work_offset_number, saved_work_offset

    if self.task == 0: # ignore the preview interpreter
        yield INTERP_EXECUTE_FINISH
        return INTERP_OK

    # ! IMPORTANT !
    #  We need to use 'yield INTERP_EXECUTE_FINISH' here to stop the read ahead 
    # and avoid it executing the rest of the remap ahead of time
    ## NOTE: No 'self.execute(..)' command can be used after 'yield INTERP_EXECUTE_FINISH'
    yield INTERP_EXECUTE_FINISH

    if twp_flag == []: # ie there is currently no TWP defined
        print("G68.4: No TWP defined to increment from. Run G68.2 first.")
        # Note the CANON_ERROR message below will not show if the command is fun in MDI
        emccanon.CANON_ERROR("G68.4: No TWP defined to increment from. Run G68.2 first.")
        yield INTERP_EXECUTE_FINISH
        yield INTERP_EXIT
        return INTERP_ERROR


    # store the current TWP to 
    twp_matrix_current = np.matrix.copy(twp_matrix)

    c = self.blocks[self.remap_level]
    p = c.p_number if c.p_flag else 0
    
    if p == 0: # true euler angles (this is the default mode)
        twp_flag = [int(p), 1, 'empty'] # one call to define the twp in this mode
        # parse requested order of rotations (default is '313' ie: ZXZ)
        q = str(int(c.q_number if c.q_flag else 313))

        if q not in ['121','131','212','232','313','323']:
            print("G68.4 (P0): No recognised Q-Word found.")
            emccanon.MESSAGE("G68.4 (P0): No recognised Q-Word found.")
            yield INTERP_EXECUTE_FINISH
            return INTERP_OK

        # parse the requested origin
        x = c.x_number if c.x_flag else 0
        y = c.y_number if c.y_flag else 0
        z = c.z_number if c.z_flag else 0
        # parse the requested rotation of tool-x around the origin
        r = c.r_number if c.r_flag else 0
        # parse requested euler angles
        th1 = c.i_number if c.i_flag else 0
        th2 = c.j_number if c.j_flag else 0
        th3 = c.k_number if c.k_flag else 0

        # build the translation vector of the twp_matrix
        twp_origin = [[x], [y], [z]]
        # convert requested origin rotation to radians 
        req_rot = r*(pi/180)
        # we use xzx-euler rotation to create the rotation matrix for the requested origin rotation
        twp_origin_rotation = twp_calc_euler_rot_matrix(0, r, 0, '131')
        print('G68.4 (P0): Twp_origin_rotation \n',twp_origin_rotation)
        # build the rotation matrix for the requested euler rotation
        twp_euler_rotation = twp_calc_euler_rot_matrix(th1, th2, th3, q)
        print('G68.4 (P0): Twp_euler_rotation \n',twp_euler_rotation)
        # calculate the total twp_rotation using matrix multiplication
        twp_rotation = np.asmatrix(twp_origin_rotation) * np.asmatrix(twp_euler_rotation)
        # combine rotation and translation and form the 4x4 twp-transformation matrix 
        twp_matrix = np.hstack((twp_rotation, twp_origin))
        twp_row_4 = [0,0,0,1]
        twp_matrix = np.vstack((twp_matrix, twp_row_4))
        twp_matrix = np.asmatrix(twp_matrix)
        # update the build state of the twp call
        twp_flag[2] = 'done'

    elif p == 1: # non-true euler angles, eg: 'pitch,roll,yaw'
        twp_flag = [int(p), 1, 'empty'] # one call to define the twp in this mode
        # parse requested order of rotations (default is '123' ie: XYZ)
        q = str(int(c.q_number if c.q_flag else 123))

        if q not in ['123','132','213','231','312','321']:
            emccanon.MESSAGE("G68.4 P1: No recognised Q-Word found.")
            yield INTERP_EXECUTE_FINISH
            return INTERP_OK

        # parse the requested origin
        x = c.x_number if c.x_flag else 0
        y = c.y_number if c.y_flag else 0
        z = c.z_number if c.z_flag else 0
        # parse the requested rotation of tool-x around the origin
        r = c.r_number if c.r_flag else 0
        req_rot = r*(pi/180)
        # parse requested angles
        th1 = c.i_number if c.i_flag else 0
        th2 = c.j_number if c.j_flag else 0
        th3 = c.k_number if c.k_flag else 0

         # build the translation vector of the twp_matrix
        twp_origin = [[x], [y], [z]]
        # convert requested origin rotation to radians 
        req_rot = r*(pi/180)
        # we use xzx-euler rotation to create the rotation matrix for the requested origin rotation
        twp_origin_rotation = twp_calc_euler_rot_matrix(0, r, 0, '131')
        print('G68.4 P1: Twp_origin_rotation \n',twp_origin_rotation)
        # build the rotation matrix for the requested euler rotation
        twp_euler_rotation = twp_calc_euler_rot_matrix(th1, th2, th3, q)
        print('G68.4 P1: Twp_euler_rotation \n',twp_euler_rotation)
        # calculate the total twp_rotation using matrix multiplication
        twp_rotation = np.asmatrix(twp_origin_rotation) * np.asmatrix(twp_euler_rotation)
        # combine rotation and translation and form the 4x4 twp-transformation matrix 
        twp_matrix = np.hstack((twp_rotation, twp_origin))
        twp_row_4 = [0,0,0,1]
        twp_matrix = np.vstack((twp_matrix, twp_row_4))
        twp_matrix = np.asmatrix(twp_matrix)
        # update the build state of the twp call
        twp_flag[2] = 'done'

    elif p == 2: # twp defined py 3 points on the plane
        # if this is the first call for this mode reset the twp_flag flag
        if not twp_flag:
            twp_flag = [int(p), 4 , 'empty', 'empty', 'empty', 'empty'] # four calls needed
            twp_build_params = {'q0':[], 'q1':[], 'q2':[], 'q3':[]}
        # Point 1: defines the origin of the twp
        # Point 2: direction from P1 to P2 defines the positive x direction on the twp (tool-x)
        # Point 3: defines the positive y side and with P1 and P2 defines the xy work plane (tool-z)
        q = int(c.q_number if c.q_flag else 0)
        # this mode needs four calls to fill all required parameters
        if q == 0: # define new origin and rotation
            x = c.x_number if c.x_flag else 0
            y = c.y_number if c.y_flag else 0
            z = c.z_number if c.z_flag else 0
            # parse the requested rotation of tool-x around the origin
            r = c.r_number if c.r_flag else 0
            req_rot = r*(pi/180)
            twp_build_params['q0'] = [x,y,z,r]
            twp_flag[2] = 'done'
        elif q == 1: # define point 1
            x1 = c.x_number if c.x_flag else 0
            y1 = c.y_number if c.y_flag else 0
            z1 = c.z_number if c.z_flag else 0
            twp_build_params['q1'] = [x1,y1,z1]
            twp_flag[3] = 'done'
        elif q == 2: # define point 2
            x2 = c.x_number if c.x_flag else 0
            y2 = c.y_number if c.y_flag else 0
            z2 = c.z_number if c.z_flag else 0
            twp_build_params['q2'] = [x2,y2,z2]
            twp_flag[4] = 'done'
        elif q == 3: # define point 3
            x3 = c.x_number if c.x_flag else 0
            y3 = c.y_number if c.y_flag else 0
            z3 = c.z_number if c.z_flag else 0
            twp_build_params['q3'] = [x3,y3,z3]
            twp_flag[5] = 'done'
        else:
            emccanon.MESSAGE("G68.4 P2: No recognised Q-Word found.")
            yield INTERP_EXECUTE_FINISH
            return INTERP_OK
        # only start calculations once all the parameters have been passed
        if twp_flag.count('done') == twp_flag[1]:
            [x, y, z] = twp_build_params['q0'][0:3]
            # build the translation vector of the twp_matrix
            twp_origin = [[x], [y], [z]]
            p1 = twp_build_params['q1'][0:3]
            p2 = twp_build_params['q2']
            p3 = twp_build_params['q3']
            print("G68.4 P2: Point 1: ",p1)
            print("G68.4 P2: Point 2: ",p2)
            print("G68.4 P2: Point 3: ",p3)
            # build vectors x:P1->P2 and v2:P1->P3
            twp_vect_x = [p2[0]-p1[0], p2[1]-p1[1], p2[2]-p1[2]]
            print("G68.4 P2: Twp_vect_x: \n",twp_vect_x)
            v2 = [p3[0]-p1[0], p3[1]-p1[1], p3[2]-p1[2]]
            print("G68.4 P2: v2",v2)
            # normalize the two vectors
            twp_vect_x = twp_vect_x / np.linalg.norm(twp_vect_x)
            v2 = v2 / np.linalg.norm(v2)
            # we can use the cross product to calculate the tool-z vector
            # note: if P3 is on the right side of the vector P1->P2
            # then the tool-z will be below the twp (ie tool-z will be downwards)
            twp_vect_z = np.cross(twp_vect_x , v2)
            print("G68.4 P2: Twp_vect_z",twp_vect_z)
            # we can use the cross product to calculate the tool-y vector
            twp_vect_y = np.cross(twp_vect_z, twp_vect_x)
            print("G68.4 P2: Twp_vect_y",twp_vect_y)
            # build the rotation matrix of the twp_matrix from the calculated tool-vectors
            # first stack the vectors (lists) and then flip diagonally (transpose)
            # so the vectors are now vertical
            twp_vect_rotation_t = np.vstack((twp_vect_x, twp_vect_y))
            twp_vect_rotation_t = np.vstack((twp_vect_rotation_t, twp_vect_z))
            twp_vect_rotation = np.transpose(twp_vect_rotation_t)
            print("G68.4 P2: Built the twp-rotation-matrix: \n", twp_vect_rotation)
            # convert requested origin rotation to radians 
            req_rot = r*(pi/180)
            # we use xzx-euler rotation to create the rotation matrix for the requested origin rotation
            twp_origin_rotation = twp_calc_euler_rot_matrix(0, r, 0, '131')
            print('G68.4 P2: Twp-origin-rotation-matrix \n',twp_origin_rotation)
            # calculate the total twp_rotation using matrix multiplication
            twp_rotation = np.asmatrix(twp_origin_rotation) * np.asmatrix(twp_vect_rotation)
            # add the origin translation on the right
            twp_matrix = np.hstack((twp_rotation, twp_origin))
            # expand to 4x4 array and make into a matrix
            twp_row_4 = [0,0,0,1]
            twp_matrix = np.vstack((twp_matrix, twp_row_4))
            twp_matrix = np.asmatrix(twp_matrix)
            print("G68.4 P2: Built twp-transformation-matrix: \n", twp_matrix)

    elif p == 3: # two vectors (vector 1 defines the tool-x and vector 2 defines the tool-z)
        q = int(c.q_number if c.q_flag else 0)
        # if this is the first call for this mode reset the twp_flag flag
        if not twp_flag:
            twp_flag = [int(p), 2 , 'empty', 'empty'] # two calls needed
            twp_build_params = {'q0':[], 'q1':[]}
        if q == 0: # define new origin and first vector (direction of x in the twp)
            x = c.x_number if c.x_flag else 0
            y = c.y_number if c.y_flag else 0
            z = c.z_number if c.z_flag else 0
            # parse the requested rotation of tool-x around the origin
            r = c.r_number if c.r_flag else 0
            req_rot = r*(pi/180)
            i = c.i_number if c.i_flag else 0
            j = c.j_number if c.j_flag else 0
            k = c.k_number if c.k_flag else 0
            twp_build_params['q0'] = [x,y,z,i,j,k]
            twp_flag[2] = 'done'
        elif q == 1: # define second vector (the normal vector of the twp
            i1 = c.i_number if c.i_flag else 0
            j1 = c.j_number if c.j_flag else 0
            k1 = c.k_number if c.k_flag else 0
            twp_build_params['q1'] = [i1,j1,k1]
            twp_flag[3] = 'done'
        else:
            emccanon.MESSAGE("G68.4 P3: No recognised Q-Word found.")
            yield INTERP_EXECUTE_FINISH
            return INTERP_OK
        # only start calculations once all the parameters have been passed
        if twp_flag.count('done') == twp_flag[1]:
            twp_origin = (x ,y, z) = twp_build_params['q0'][0:3]
            (i, j, k) = twp_build_params['q0'][3:6]
            (i1, j1, k1) = twp_build_params['q1']
            print("x, y, z", (x, y, z))
            print("i, j, k", (i, j, k))
            print("i1, j1, k1", (i1, j1, k1))
            # build unit vector defining tool-x direction
            twp_vect_x = [i-x, j-y, k-z]
            twp_vect_x = twp_vect_x / np.linalg.norm(twp_vect_x)
            twp_vect_z = [i1, j1, k1]
            twp_vect_z = twp_vect_z / np.linalg.norm(twp_vect_z)
            orth = np.dot(twp_vect_x, twp_vect_z)
            print("orth check: ", orth)
            # the two vectors must be orthogonal
            if orth != 0:
                emccanon.MESSAGE("G68.4 P3: Vectors are not orthogonal.")
                # reset the parameter values
                twp_flag = [int(p), 2 , 'empty', 'empty'] # two calls needed
                twp_build_params = {'q0':[], 'q1':[]}
                yield INTERP_EXECUTE_FINISH
                return INTERP_OK
            # we can use the cross product to calculate the tool-y vector
            twp_vect_y = np.cross(twp_vect_z, twp_vect_x)
            print("G68.4 P3: twp_vect_y",twp_vect_y)
            # build the rotation matrix of the twp_matrix from the calculated tool-vectors
            # first stack the vectors (lists) and then flip diagonally (transpose)
            # so the vectors are now vertical
            twp_vect_rotation_t = np.vstack((twp_vect_x, twp_vect_y))
            twp_vect_rotation_t = np.vstack((twp_vect_rotation_t, twp_vect_z))
            twp_vect_rotation = np.transpose(twp_vect_rotation_t)
            print("G68.4 P3: Built twp-rotation-matrix: \n", twp_vect_rotation)
            # convert requested origin rotation to radians 
            req_rot = r*(pi/180)
            # we use xzx-euler rotation to create the rotation matrix for the requested origin rotation
            twp_origin_rotation = twp_calc_euler_rot_matrix(0, r, 0, '131')
            print('G68.4 P3: Twp-origin-rotation-matrix \n',twp_origin_rotation)
            # calculate the total twp_rotation using matrix multiplication
            twp_rotation = np.asmatrix(twp_origin_rotation) * np.asmatrix(twp_vect_rotation)
            # add the origin translation on the right
            twp_origin = [[x], [y], [z]]
            twp_matrix = np.hstack((twp_rotation, twp_origin))
            # expand to 4x4 array and make into a matrix
            twp_row_4 = [0,0,0,1]
            twp_matrix = np.vstack((twp_matrix, twp_row_4))
            twp_matrix = np.asmatrix(twp_matrix)
            print("G68.4 P3: Built twp-transformation-matrix: \n", twp_matrix)

    else:
        emccanon.MESSAGE("G68.4: No recognised P-Word found.")
        yield INTERP_EXECUTE_FINISH
        return INTERP_OK

    print("G68.4: twp_flag: ", twp_flag)
    print("G68.4: calls required: ", twp_flag.count('done'))
    print("G68.4: number of calls made: ", twp_flag.count('done'))

    if twp_flag.count('done') == twp_flag[1]:
        print('G68.4: requested rotation',req_rot)
        print("G68.4: twp_matrix_current: \n", twp_matrix_current)
        print("G68.4: incremental twp_matrix requested: \n",twp_matrix)
        print("G68.4: calculating new twp_matrix...")
        twp_matrix_new = twp_matrix_current * twp_matrix
        print("G68.4: twp_matrix_new: \n",twp_matrix_new)
        twp_origin = [twp_matrix[0,3],twp_matrix[1,3],twp_matrix[2,3]]
        print("G68.4: twp origin: ", twp_origin)
        twp_vect_x = [twp_matrix[0,0],twp_matrix[1,0],twp_matrix[2,0]]
        print("G68.4: twp vector-x: ", twp_vect_x)
        twp_vect_z = [twp_matrix[0,2],twp_matrix[1,2],twp_matrix[2,2]]
        print("G68.4: twp vector-z: ", twp_vect_z)
        print("G68.4: incremented twp_matrix: \n", twp_matrix_new)
        twp_matrix = twp_matrix_new
        hal.set_p("twp_helper_comp.twp-is-defined-in",str(1))
        yield INTERP_EXECUTE_FINISH

        gui_update_twp(self)
    return INTERP_OK

